"""This module contains classes that provide nearly identical functionality
as built-in Python :py:class:`list` objects, but which require that all items
in a list be of a specified type
"""

from .typedlist import TypedList, TypedListWithID
