"""**Code for processing and managing numerical types of data**

The :py:mod:`pyxx.numbers` module contains a variety of code intended to aid
in processing numerical data.  For instance, this module contains code to
identify whether a value is a valid floating-point number or integer.
"""

from .functions import is_float, is_integer
