"""Collection of utilities for Python code"""

# PROGRAM VERSION ------------------------------------------------------------
__version__ = '1.1.0'


# PACKAGE MODULES ------------------------------------------------------------
from . import arrays
from . import dev
from . import files
from . import numbers
from . import strings
from . import units
