"""This module contains classes that represent various types of files (binary
files, text files, etc.) and can be used to perform general file operations)
"""

from .binaryfile import BinaryFile
from .file import File
from .textfile import TextFile
